import os
from typing import Union

import numpy as np
import tifffile
import torch

from ..multi_output_unet3d.multi_output_unet3d import MultiOutputUnet3D
from ..progress import ProgressNotifier
from ..utils import get_device


class Predict:
    """Class for prediction of volumetric (3D) data using a 3D U-Net model."""

    def __init__(self, imgs, model_params, result_path=None,
                 network=MultiOutputUnet3D,
                 max_patch_size=(64, 256, 256),  # (depth, height, width)
                 overlap_factor=0.1,
                 batch_size=1,
                 normalization_mode='single',
                 clip_threshold=(0., 99.98),
                 add_tile=0,
                 compress_tif=False,
                 show_progress=True,
                 device: Union[torch.device, str] = 'auto',
                 progress_notifier: ProgressNotifier = ProgressNotifier.progress_notifier_tqdm()):

        if device == 'auto':
            self.device = get_device()
        else:
            self.device = torch.device(device)

        if isinstance(imgs, str):
            imgs = tifffile.imread(imgs)

        self.max_patch_size = max_patch_size
        self.overlap_factor = overlap_factor
        self.batch_size = batch_size
        self.add_tile = add_tile
        self.normalization_mode = normalization_mode
        self.clip_threshold = clip_threshold
        self.result_path = result_path
        self.compress_tif = compress_tif
        self.show_progress = show_progress

        # Preprocess and split data into patches
        imgs = imgs.astype('float32')
        imgs = self.__reshape_data(imgs)
        imgs = self.__preprocess(imgs)
        patches = self.__split(imgs)
        del imgs

        # Load model parameters and initialize model
        self.model_params = torch.load(model_params, map_location=self.device)
        self.model = network(in_channels=self.model_params['in_channels'],
                             n_filter=self.model_params['n_filter'],
                             output_heads=self.model_params['output_heads'],
                             use_interpolation=self.model_params.get('use_interpolation', True)).to(self.device)

        self.model.load_state_dict(self.model_params['state_dict'])
        self.model.eval()

        self.target_keys = list(self.model_params['output_heads'].keys())

        # Predict patches
        result_patches = self.__predict(patches, progress_notifier)
        del patches, self.model

        # Stitch patches back into full volume
        result = self.__stitch(result_patches)
        del result_patches, self.model_params

        # Save or store results
        if self.result_path is not None:
            for target_key in self.target_keys:
                target_file = self.result_path + target_key + '.tif' if os.path.exists(self.result_path) else (
                        self.result_path + '_' + target_key + '.tif')
                tifffile.imwrite(target_file,
                                 result[target_key],
                                 compression='deflate' if self.compress_tif else None)
            del result
            self.result = None
        else:
            self.result = result

        if torch.cuda.is_available():
            torch.cuda.empty_cache()

    def __reshape_data(self, imgs):
        if imgs.ndim == 3:  # Single volume without channel dimension (D,H,W)
            imgs = np.expand_dims(imgs, axis=0)  # Add batch dimension: (1,D,H,W)

        elif imgs.ndim == 4:  # Multiple volumes (N,D,H,W), already correct shape
            pass

        else:
            raise ValueError(f"Unsupported input shape: {imgs.shape}")

        self.imgs_shape = imgs.shape  # (N,D,H,W)

        return imgs

    def __preprocess(self, imgs):
        if self.normalization_mode == 'single':
            for i in range(len(imgs)):
                img = imgs[i]
                img_clipped = np.clip(img,
                                      np.percentile(img, self.clip_threshold[0]),
                                      np.percentile(img, self.clip_threshold[1]))
                img_normed = (img_clipped - img_clipped.min()) / (img_clipped.ptp() + 1e-8)
                imgs[i] = img_normed

        elif self.normalization_mode in ['first', 'all']:
            reference_img = imgs[0] if self.normalization_mode == 'first' else imgs
            clip_min, clip_max = np.percentile(reference_img, [self.clip_threshold[0],
                                                               self.clip_threshold[1]])
            imgs_clipped = np.clip(imgs, clip_min, clip_max)
            imgs_normed = (imgs_clipped - clip_min) / (clip_max - clip_min + 1e-8)
            imgs[:] = imgs_normed

        else:
            raise ValueError(f'Invalid normalization mode: {self.normalization_mode}')

        return imgs

    def __split(self, imgs):
        N_volumes, D_img, H_img, W_img = imgs.shape
        D_patch, H_patch, W_patch = [
            min(s_img, s_patch) for s_img, s_patch in zip((D_img, H_img, W_img), self.max_patch_size)
        ]

        # Define strides
        stride_dhw = [max(1, int(s * (1 - self.overlap_factor))) for s in (D_patch, H_patch, W_patch)]

        # Calculate starting indices for each dimension
        self.Z_start = list(range(0, max(D_img - D_patch + 1, 1), stride_dhw[0]))
        if self.Z_start[-1] + D_patch < D_img:
            self.Z_start.append(D_img - D_patch)

        self.Y_start = list(range(0, max(H_img - H_patch + 1, 1), stride_dhw[1]))
        if self.Y_start[-1] + H_patch < H_img:
            self.Y_start.append(H_img - H_patch)

        self.X_start = list(range(0, max(W_img - W_patch + 1, 1), stride_dhw[2]))
        if self.X_start[-1] + W_patch < W_img:
            self.X_start.append(W_img - W_patch)

        # Store patch size and counts as attributes for stitching
        self.patch_size = (D_patch, H_patch, W_patch)
        self.N_z = len(self.Z_start)
        self.N_y = len(self.Y_start)
        self.N_x = len(self.X_start)
        self.N_per_vol = self.N_z * self.N_y * self.N_x
        self.imgs_shape = imgs.shape

        patches_list = []

        # Extract patches
        for vol in range(N_volumes):
            for z in self.Z_start:
                for y in self.Y_start:
                    for x in self.X_start:
                        patch = imgs[
                                vol,
                                z: z + D_patch,
                                y: y + H_patch,
                                x: x + W_patch
                                ]
                        patches_list.append(patch)

        patches = np.stack(patches_list)[:, None, ...]  # (N_patches_total, C=1, D,H,W)

        return patches

    def __predict(self, patches, progress_notifier=None):

        results = {k: [] for k in self.target_keys}

        with torch.no_grad():
            num_batches = int(np.ceil(len(patches) / self.batch_size))

            iterator = range(num_batches)

            if progress_notifier and self.show_progress:
                iterator = progress_notifier.iterator(iterator)

            for idx in iterator:
                batch_slice = slice(idx * self.batch_size, (idx + 1) * self.batch_size)
                batch = torch.tensor(patches[batch_slice], dtype=torch.float32).to(self.device)

                preds = self.model(batch)

                for key in results.keys():
                    results[key].append(preds[key].cpu().numpy())

                del batch, preds

        results = {key: np.concatenate(results[key]) for key in results}

        return results

    def __stitch(self, result_patches, blend_margin=16):
        """
        Stitch 3D patches back into a complete volume with proper overlap handling.

        Parameters:
        -----------
        result_patches : dict
            Dictionary of patches for each output head
        blend_margin : int
            Size of the blending margin for smooth transitions

        Returns:
        --------
        dict
            Dictionary containing stitched volumes for each output head
        """
        result = {}

        # Get dimensions from the original volume shape
        n_volumes, depth, height, width = self.imgs_shape

        for target_key in self.target_keys:
            # Get output channels for this target
            n_channels = self.model_params['output_heads'][target_key]['channels']

            # Initialize result volume and weight map
            result_volume = np.zeros(
                (n_volumes, n_channels, depth, height, width),
                dtype='float32'
            )
            weight_map = np.zeros_like(result_volume)

            # Process each volume separately
            for vol_idx in range(n_volumes):
                # Get patches for this volume
                vol_patches = result_patches[target_key][
                              vol_idx * self.N_per_vol:(vol_idx + 1) * self.N_per_vol
                              ].reshape(self.N_z, self.N_y, self.N_x, n_channels, *self.patch_size)

                # Iterate through all patches
                for z_idx, z_start in enumerate(self.Z_start):
                    for y_idx, y_start in enumerate(self.Y_start):
                        for x_idx, x_start in enumerate(self.X_start):
                            # Get current patch
                            patch = vol_patches[z_idx, y_idx, x_idx]

                            # Create weight mask with smooth transitions at borders
                            patch_weight = np.ones_like(patch)

                            # Apply blending weights at patch edges where overlaps exist
                            if z_idx > 0:  # Front overlap
                                for i in range(min(blend_margin, self.N_z)):
                                    patch_weight[:, i, :, :] = i / blend_margin
                            if z_idx < self.N_z - 1:  # Back overlap
                                for i in range(min(blend_margin, self.N_z)):
                                    patch_weight[:, max(-(i + 1), 0), :, :] = i / blend_margin

                            if y_idx > 0:  # Top overlap
                                for i in range(blend_margin):
                                    patch_weight[:, :, i, :] = i / blend_margin
                            if y_idx < self.N_y - 1:  # Bottom overlap
                                for i in range(blend_margin):
                                    patch_weight[:, :, max(-(i + 1), 0), :] = i / blend_margin

                            if x_idx > 0:  # Left overlap
                                for i in range(blend_margin):
                                    patch_weight[:, :, :, i] = i / blend_margin
                            if x_idx < self.N_x - 1:  # Right overlap
                                for i in range(blend_margin):
                                    patch_weight[:, :, :, max(-(i + 1), 0)] = i / blend_margin

                            # Define slices for placing the patch in the result volume
                            z_slice = slice(z_start, min(z_start + self.patch_size[0], depth))
                            y_slice = slice(y_start, min(y_start + self.patch_size[1], height))
                            x_slice = slice(x_start, min(x_start + self.patch_size[2], width))

                            # Get corresponding patch slice (in case patch extends beyond volume)
                            patch_z_slice = slice(0, z_slice.stop - z_slice.start)
                            patch_y_slice = slice(0, y_slice.stop - y_slice.start)
                            patch_x_slice = slice(0, x_slice.stop - x_slice.start)

                            # Add weighted patch contribution to result
                            result_volume[vol_idx, :, z_slice, y_slice, x_slice] += (
                                    patch[:, patch_z_slice, patch_y_slice, patch_x_slice] *
                                    patch_weight[:, patch_z_slice, patch_y_slice, patch_x_slice]
                            )

                            # Add weights to weight map
                            weight_map[vol_idx, :, z_slice, y_slice, x_slice] += (
                                patch_weight[:, patch_z_slice, patch_y_slice, patch_x_slice]
                            )

            # Normalize by weights to get final result
            # Avoid division by zero
            mask = weight_map > 0
            result_volume[mask] = result_volume[mask] / weight_map[mask]

            # Handle any unweighted areas (should be rare)
            if np.any(~mask):
                result_volume[~mask] = 0

            # Remove singleton dimensions if necessary
            result[target_key] = np.squeeze(result_volume)

        return result



import os
import shutil
from pathlib import Path
from typing import Tuple, List

import numpy as np
import tifffile
import torch

from albumentations import (ShiftScaleRotate, GaussNoise, ShotNoise, RandomCrop3D, RandomBrightnessContrast, Compose,
                            Blur)
from torch.utils.data import Dataset


class DataProcess(Dataset):
    """
    Class for training data generation (processing, splitting, augmentation)

    Create training data object for network training

    1) Create folder structure for training data
    2) Move and preprocess training volumes
    3) Split input volumes into patches
    4) Augment training data
    5) Create object of PyTorch Dataset class for training

    Parameters
    ----------
    source_dir : Tuple[str, str]
        Path of training data [volumes, labels]. volumes need to be tif files.
    dim_out : Tuple[int, int, int]
        Resize dimensions of volumes for training
    data_path : str
        Base path of temporary directories for training data
    clip_threshold : Tuple[float, float]
        Clip thresholds for intensity normalization of volumes
    add_patch : int
        Add additional patch for splitting volumes into patches with more overlapping patches
    val_split : float
        Validation split for training
    aug_factor : int
        Factor of volume augmentation
    shiftscalerotate : Tuple[float, float, float]
        Shift, scale and rotate during augmentation
    gauss_noise_lims : Tuple[float, float]
        Limits of Gaussian noise for volume augmentation
    shot_noise_lims : Tuple[float, float]
        Limits of Shot noise for volume augmentation
    brightness_contrast : Tuple[float, float]
        Adapt brightness and contrast of volumes during augmentation
    blur_limit : Tuple[int, int]
        Blur limit range for augmentation
    create : bool
        If False, existing data set in data_path is used
    """

    def __init__(self,
                 volume_dir: str,
                 target_dirs: List[str],
                 data_dir: str = '../data/',
                 dim_out: Tuple[int, int, int] = (128, 128, 128),
                 in_channels: int = 1,
                 add_tile: int = 0,
                 nan_to_val: float = 0,
                 val_split: float = 0.2,
                 clip_threshold: Tuple[float, float] = (0., 99.99),
                 aug_factor: int = 10,
                 scale_limit: Tuple[float, float] = (-0.75, 0),
                 rotate_limit: Tuple[float, float] = (0, 360),
                 gauss_noise_lims: Tuple[float, float] = (0.01, 0.1),
                 shot_noise_lims: Tuple[float, float] = (0.005, 0.01),
                 brightness_contrast: Tuple[float, float] = (0.1, 0.1),
                 blur_limit: Tuple[int, int] = (3, 7),
                 random_rotate: bool = True,
                 create: bool = True):

        self.volume_dir = volume_dir
        self.target_dirs = target_dirs
        self.target_keys = [os.path.basename(os.path.normpath(_dir)) for _dir in target_dirs]
        self.data_dir = data_dir
        self.data = []  # empty list for data dicts
        self.create = create
        self.in_channels = in_channels
        self.dim_out = dim_out
        self.nan_to_val = nan_to_val
        self.clip_threshold = clip_threshold
        self.add_tile = add_tile
        self.aug_factor = aug_factor
        self.scale_limit = scale_limit
        self.rotate_limit = rotate_limit
        self.brightness_contrast = brightness_contrast
        self.gauss_noise_lims = gauss_noise_lims
        self.shot_noise_lims = shot_noise_lims
        self.blur_limit = blur_limit
        self.random_rotate = random_rotate

        self.val_split = val_split
        self.mode = 'train'

        if create:
            # Delete the existing data directory if it exists
            if os.path.exists(self.data_dir):
                shutil.rmtree(self.data_dir)

            # Create a new data directory
            os.makedirs(self.data_dir, exist_ok=True)

            # Proceed with data processing
            self.__read_and_edit()
            if self.aug_factor is not None:
                self.__augment()

    def __read_and_edit(self):
        # Find all TIFF files in volume_dir
        volume_path = Path(self.volume_dir)
        files_volume = [str(file) for ext in ['*.tif', '*.tiff'] for file in volume_path.glob(ext)]

        # Read and normalize volumes
        for file_img_i in files_volume:
            img_i = tifffile.imread(file_img_i)
            basename_i = os.path.basename(file_img_i)

            # Create dict for volume and targets
            data_i = {}

            # Clip and normalize
            img_i = img_i.astype(np.float32)
            img_i = np.clip(img_i, a_min=np.nanpercentile(img_i, self.clip_threshold[0]),
                            a_max=np.percentile(img_i, self.clip_threshold[1]))
            img_i = (img_i - np.nanmin(img_i)) / (np.nanmax(img_i) - np.nanmin(img_i))

            # Add volume to dict
            data_i['volume'] = img_i

            # Iterate through target folders and add targets to dict
            all_targets_exist = True
            for target_dir_j in self.target_dirs:
                file_target_j = os.path.join(target_dir_j, basename_i)
                if os.path.exists(file_target_j):
                    target_data = tifffile.imread(file_target_j)
                    target_key = os.path.basename(target_dir_j[:-1])  # Use the folder name as the target key
                    data_i[target_key] = target_data
                else:
                    print(f"Warning: Target file {file_target_j} does not exist.")
                    all_targets_exist = False
                    break

            # Add data to the list only if all targets exist
            if all_targets_exist:
                self.data.append(data_i)

    def __augment(self, p=0.8):
        # Define additional targets for masks
        additional_targets_3D = {key: 'mask3d' for key in self.target_keys}

        # Pipeline for 3D spatial transformations
        aug_pipeline_3D = Compose(
            transforms=[
                ShiftScaleRotate(shift_limit=0, scale_limit=self.scale_limit, rotate_limit=self.rotate_limit, p=0.8),
                RandomCrop3D(size=self.dim_out),
            ],
            additional_targets=additional_targets_3D
        )

        # Pipeline for 2D intensity transformations (slice-by-slice)
        aug_pipeline = Compose(
            transforms=[
                RandomBrightnessContrast(
                    brightness_limit=self.brightness_contrast[0],
                    contrast_limit=self.brightness_contrast[1],
                    p=0.5
                ),
                Blur(blur_limit=self.blur_limit, p=0.3),
                ShotNoise(scale_range=self.shot_noise_lims, p=0.5),
                GaussNoise(std_range=self.gauss_noise_lims, p=0.5),
            ],
            p=p, additional_targets=additional_targets_3D,
        )

        running_number = 0

        # Iterate over data patches
        for patch_data in self.data:
            # Extract main volume and target masks
            volume_i = patch_data['volume']
            targets_i = {key: patch_data[key] for key in patch_data if key != 'volume'}

            # Prepare input data for 3D pipeline
            data_i = {'volume': volume_i}
            data_i.update(targets_i)

            for _ in range(self.aug_factor):
                # Apply 3D spatial transformations
                augmented_3D = aug_pipeline_3D(**data_i)

                # Extract augmented volume and targets from the first pipeline
                _aug_volume = augmented_3D['volume']
                _aug_targets = {key: augmented_3D[key] for key in targets_i}

                # Prepare input data for 2D pipeline (remap "volume" to "image")
                data_j = {'images': _aug_volume}
                data_j.update(_aug_targets)

                # Apply 2D intensity transformations
                augmented = aug_pipeline(**data_j)

                # Extract final augmented volume and targets
                aug_volume = augmented['images']
                aug_targets = {key: augmented[key] for key in targets_i}

                # Save augmented volume
                volume_dir = os.path.join(self.data_dir, 'volume')
                os.makedirs(volume_dir, exist_ok=True)
                tifffile.imwrite(
                    os.path.join(volume_dir, f'volume_{running_number}.tif'),
                    aug_volume
                )

                # Save augmented masks/targets
                for target_key, target_value in aug_targets.items():
                    target_dir = os.path.join(self.data_dir, target_key)
                    os.makedirs(target_dir, exist_ok=True)
                    tifffile.imwrite(
                        os.path.join(target_dir, f'{target_key}_{running_number}.tif'),
                        target_value
                    )

                running_number += 1

        print(f'Augmentation completed for {running_number} patches.')

    def __len__(self):
        # Count the number of volume files in the 'volume' directory within the data path
        volume_dir = os.path.join(self.data_dir, 'volume')
        return len([name for name in os.listdir(volume_dir) if os.path.isfile(os.path.join(volume_dir, name))])

    def __getitem__(self, idx):
        # Construct the file name for the volume
        img_name = f'volume_{idx}.tif'

        # Read the volume from the 'volume' directory
        volume_path = os.path.join(self.data_dir, 'volume', img_name)
        volume_0 = tifffile.imread(volume_path)

        # Convert the volume to a PyTorch tensor
        volume = torch.from_numpy(volume_0)

        # Initialize the sample dictionary with the volume
        sample = {'volume': volume}

        # Iterate over each target directory and read the corresponding target file
        for target_dir in self.target_dirs:
            target_key = os.path.basename(os.path.normpath(target_dir))
            target_file = f'{target_key}_{idx}.tif'
            target_path = os.path.join(self.data_dir, target_key, target_file)
            if os.path.exists(target_path):
                target_data = tifffile.imread(target_path).astype('float32')
                if target_key == 'orientation':
                    target_data = np.stack([np.cos(target_data), np.sin(target_data)])
                # set NaNs to val
                target_data[np.isnan(target_data)] = self.nan_to_val
                sample[target_key] = torch.from_numpy(target_data)
            else:
                raise FileNotFoundError(f"Target file {target_path} not found.")

        # Return the sample as a dictionary
        return sample



import cv2
from os import listdir
from os.path import isfile, join
import numpy as np

def threshold_images(in_path, out_path):
    """
    Performs the threshold function on all the images in the folder `in_path` and outputs them in `out_path`
    Params:
        in_path: folder of source images
        out_path: folder to place images
    """

    files = [f for f in listdir(in_path) if isfile(join(in_path, f))]
    for f in files:
        img = np.array(cv2.imread(join(in_path, f), cv2.IMREAD_GRAYSCALE))
        out_img = ((img >= 150) * np.ones(img.shape) * 255 ).astype(np.uint8)
        cv2.imwrite(filename=join(out_path, f), img=out_img, )

def invert_images(in_path, out_path):
    """
    Performs the invert function on all the images in the folder `in_path` and outputs them in `out_path`
    Params:
        in_path: folder of source images
        out_path: folder to place images
    """

    files = [f for f in listdir(in_path) if isfile(join(in_path, f))]
    for f in files:
        img = np.array(cv2.imread(join(in_path, f), cv2.IMREAD_GRAYSCALE))
        img = (np.ones(img.shape) * 255 - img).astype(np.uint8)
        cv2.imwrite(filename=join(out_path, f), img=img, )


if __name__ == '__main__':
    # threshold all the images I have drawn 
    pass


try:
    import torch
except ImportError:
    raise ImportError  # Torch is not installed. Please install through https://pytorch.org/get-started/locally/

from .data import DataProcess
from .train import Trainer
from .predict import Predict
from .multi_output_unet3d import MultiOutputUnet3D
from .losses import *



import torch

def get_gpu_name():
	return torch.cuda.get_device_name(0)

def get_all_gpu_names():
	d = []
	for i in range(torch.cuda.device_count()):
		d.append(torch.cuda.get_device_name(i))
	return d

if __name__ == '__main__':
	print(get_all_gpu_names())


import os
import time
from typing import Union

import torch.optim as optim
from matplotlib import pyplot as plt
from torch.utils.data import DataLoader, random_split
from tqdm import tqdm

from .losses import *
from .multi_output_unet3d import MultiOutputUnet3D
from ..utils import init_weights, get_device


class Trainer:
    """
    Class for training of neural network. Creates trainer object, training is started with .start().

    Parameters
    ----------
    dataset
        Training data, object of PyTorch Dataset class
    output_heads
        Dictionary of output heads.
    num_epochs : int
        Number of training epochs
    network
        Network class (Default Unet)
    use_interpolation : bool
        Whether to use interpolation in decoder instead of transpose convolution.
    batch_size : int
        Batch size for training
    lr : float
        Learning rate
    in_channels : int
        Number of input channels
    n_filter : int
        Number of convolutional filters in first layer
    val_split : float
        Validation split
    save_dir : str
        Path of directory to save trained networks
    save_name : str
        Base name for saving trained networks
    save_iter : bool
        If True, network state is save after each epoch
    load_weights : str, optional
        If not None, network state is loaded before training
    loss_function : str
        Loss function ('BCEDice', 'Tversky' or 'logcoshTversky')
    loss_params : Tuple[float, float]
        Parameter of loss function, depends on chosen loss function
    device : torch.device or str, optional
        Device to run the pytorch model on, defaults to 'auto', which selects CUDA or MPS if available.
    """

    def __init__(self, dataset, output_heads, num_epochs, network=MultiOutputUnet3D, use_interpolation=False,
                 batch_size=4, lr=1e-3, in_channels=1, n_filter=64, dilation=1, val_split=0.2,
                 save_dir='./', save_name='model.pt', save_iter=False, load_weights=False, loss_function='BCEDice',
                 loss_params=(0.5, 0.5), time_loss_weight=0.1, device: Union[torch.device, str] = 'auto'):

        if device == 'auto':
            self.device = get_device()
        else:
            self.device = torch.device(device)
        self.network = network
        self.model = network(n_filter=n_filter, in_channels=in_channels, output_heads=output_heads,
                             use_interpolation=use_interpolation).to(self.device)
        self.model.apply(init_weights)
        self.data = dataset
        self.output_heads = output_heads
        self.num_epochs = num_epochs
        self.batch_size = batch_size
        self.lr = lr
        self.best_loss = torch.tensor(float('inf'))
        self.save_iter = save_iter
        self.loss_function = loss_function
        self.loss_params = loss_params
        self.time_loss_weight = time_loss_weight
        self.n_filter = n_filter
        self.in_channels = in_channels
        self.use_interpolation = use_interpolation

        # Define loss functions for each head
        self.loss_functions = {
            name: self._get_loss_function(config['loss']) for name, config in self.output_heads.items()
        }

        # Define activations for each head
        self.activations = {
            name: config.get('activation', None) for name, config in self.output_heads.items()
        }

        # Define weights for each head
        self.loss_weights = {
            name: config.get('weight', 1.0) for name, config in self.output_heads.items()
        }

        # split training and validation data
        num_val = int(len(dataset) * val_split)
        num_train = len(dataset) - num_val
        self.dim = dataset.dim_out
        train_data, val_data = random_split(dataset, [num_train, num_val])
        self.train_loader = DataLoader(train_data, batch_size=self.batch_size, pin_memory=True, drop_last=True)
        self.val_loader = DataLoader(val_data, batch_size=self.batch_size, pin_memory=True, drop_last=True)
        if loss_function == 'BCEDice':
            self.criterion = BCEDiceLoss(loss_params[0], loss_params[1])
        elif loss_function == 'Tversky':
            self.criterion = TverskyLoss(loss_params[0], loss_params[1])
        elif loss_function == 'logcoshTversky':
            self.criterion = logcoshTverskyLoss(loss_params[0], loss_params[1])
        elif loss_function == 'BCEDiceTemporalLoss':
            self.criterion = BCEDiceTemporalLoss(loss_params=loss_params)
        else:
            raise ValueError(f'Loss "{loss_function}" not defined!')
        self.criterion_time = nn.SmoothL1Loss()
        self.optimizer = torch.optim.Adam(self.model.parameters(), lr=self.lr)
        self.scheduler = optim.lr_scheduler.ReduceLROnPlateau(self.optimizer, mode='min', patience=5, factor=0.2)
        self.save_dir = save_dir
        os.makedirs(self.save_dir, exist_ok=True)
        self.save_name = save_name
        self.params = {'optimizer': self.optimizer.state_dict(),
                       'lr': self.lr,
                       'loss_function': self.loss_function,
                       'loss_params': self.loss_params,
                       'time_loss_weight': self.time_loss_weight,
                       'n_filter': self.n_filter,
                       'use_interpolation': use_interpolation,
                       'dilation': dilation, 'batch_size': self.batch_size,
                       'augmentation': self.data.aug_factor,
                       'clip_threshold': self.data.clip_threshold,
                       'scale_limit': self.data.scale_limit,
                       'rotate_limit': self.data.rotate_limit,
                       'gauss_noise_lims': self.data.gauss_noise_lims,
                       'shot_noise_lims': self.data.shot_noise_lims,
                       'blur_limit': self.data.blur_limit,
                       'random_rotate': self.data.random_rotate,
                       'brightness_contrast': self.data.brightness_contrast,
                       'in_channels': in_channels,
                       'output_heads': output_heads}
        if load_weights:
            self.state = torch.load(os.path.join(self.save_dir, self.save_name))
            self.model.load_state_dict(self.state['state_dict'])
            self.epoch_start = self.state['epoch']
        else:
            self.epoch_start = 0

    @staticmethod
    def _get_loss_function(loss_name):
        if loss_name == 'BCEDiceLoss':
            return BCEDiceLoss(1, 1)
        elif loss_name == 'DiceLoss':
            return BCEDiceLoss(0, 1)
        elif loss_name == 'TverskyLoss':
            return TverskyLoss()
        elif loss_name == 'logcoshTverskyLoss':
            return logcoshTverskyLoss()
        elif loss_name == 'BCEDiceTemporalLoss':
            return BCEDiceTemporalLoss()
        else:
            raise ValueError(f'Loss "{loss_name}" not defined!')

    @staticmethod
    def _apply_activation(x, activation):
        if activation == 'sigmoid':
            return torch.sigmoid(x)
        elif activation == 'tanh':
            return torch.tanh(x)
        elif activation == 'relu':
            return torch.relu(x)
        elif activation == 'softmax':
            return torch.softmax(x, dim=1)
        return x

    def __iterate(self, epoch, mode):
        if mode == 'train':
            running_loss = 0.0
            for i, batch_i in tqdm(enumerate(self.train_loader), total=len(self.train_loader), unit='batch'):
                x_i = batch_i['volume'].to(self.device, non_blocking=True)
                y_i = {key: batch_i[key].to(self.device, non_blocking=True) for key in self.output_heads}

                if x_i.dim() == 4:
                    x_i = x_i.unsqueeze(1)

                # Forward pass
                y_pred = self.model(x_i)

                # Handle outputs
                total_loss = 0
                for name, config in self.output_heads.items():
                    target = y_i[name]
                    if target.dim() == 4:
                        target = target.unsqueeze(1)
                    pred = y_pred[name]
                    loss = self.loss_functions[name](pred, target)
                    total_loss += self.loss_weights[name] * loss

                # Backward pass and optimization
                self.optimizer.zero_grad()
                total_loss.backward()
                torch.nn.utils.clip_grad_norm_(self.model.parameters(), max_norm=1.0)
                self.optimizer.step()

        elif mode == 'val':
            loss_list = []
            with torch.no_grad():
                for i, batch_i in tqdm(enumerate(self.val_loader), total=len(self.val_loader), unit='batch'):
                    x_i = batch_i['volume'].to(self.device, non_blocking=True)
                    y_i = {key: batch_i[key].to(self.device, non_blocking=True) for key in self.output_heads}

                    if x_i.dim() == 4:
                        x_i = x_i.unsqueeze(1)

                    # Forward pass
                    y_pred = self.model(x_i)

                    # Compute combined loss
                    total_loss = 0
                    for name in self.output_heads:
                        target = y_i[name]

                        if target.dim() == 4:
                            target = target.unsqueeze(1)
                        pred = self._apply_activation(y_pred[name], self.activations.get(name))
                        loss = self.loss_functions[name](pred, target)
                        total_loss += self.loss_weights[name] * loss

                    loss_list.append(total_loss.detach())

            val_loss = torch.stack(loss_list).mean()
            return val_loss

        torch.cuda.empty_cache()

    def start(self):
        for epoch in range(self.num_epochs):
            start_time = time.time()  # Start time of epoch

            # Training phase
            self.__iterate(epoch, 'train')

            # Update state dictionary
            self.state = {
                'epoch': epoch + self.epoch_start,
                'epoch_start': self.epoch_start,
                'best_loss': self.best_loss,
                'state_dict': self.model.state_dict()
            }
            self.state.update(self.params)

            # Validation phase
            with torch.no_grad():
                val_loss = self.__iterate(epoch, 'val')
                self.scheduler.step(val_loss)

            # Calculate time taken for epoch
            end_time = time.time()
            time_taken = round(end_time - start_time, 2)

            print(f"\nEpoch {epoch} completed in {time_taken} seconds.")

            # Check for improvement in validation loss
            if val_loss < self.best_loss:
                print(
                    f'\nValidation loss improved from {self.best_loss.item():.5f} to {val_loss.item():.5f} - saving model state')
                self.state['best_loss'] = self.best_loss = val_loss
                torch.save(self.state, os.path.join(self.save_dir, self.save_name))
            else:
                print(f'\nValidation loss did not improve from {self.best_loss.item():.5f}')

            # Save model state after each epoch if required
            if self.save_iter:
                torch.save(self.state, os.path.join(self.save_dir, f'model_epoch_{epoch + self.epoch_start}.pt'))


import torch
from torch import nn
from typing import Dict, List, Optional


class MultiOutputUnet(nn.Module):
    def __init__(self, in_channels=1, output_heads: Dict[str, dict] = None, n_filter=32, **kwargs):
        """
        Multi-output U-Net architecture supporting various output heads.

        Parameters
        ----------
        in_channels : int
            Number of input channels
        output_heads : Dict[str, dict]
            Dictionary defining output heads, e.g.,
            {
                'target1': {'channels': 1, 'activation': 'sigmoid', 'loss': 'BCEDice', 'weight': 0.5},
                'target2': {'channels': 1, 'activation': 'ReLU', 'loss': 'BCEDice'},
                'target3': {'channels': 2, 'activation': None, 'loss': 'MSE'}
            }
        n_filter : int
            Number of base convolutional filters
        """
        super().__init__()
        self.output_heads = output_heads or {
            'default': {'channels': 1, 'activation': 'sigmoid'}
        }
        self.deep_supervision = False

        # Encoder
        self.encode1 = self.conv(in_channels=in_channels, out_channels=n_filter)
        self.encode2 = self.conv(n_filter, n_filter)
        self.maxpool1 = nn.MaxPool2d(kernel_size=2, stride=2)
        self.encode3 = self.conv(n_filter, 2 * n_filter)
        self.encode4 = self.conv(2 * n_filter, 2 * n_filter)
        self.maxpool2 = nn.MaxPool2d(kernel_size=2, stride=2)
        self.encode5 = self.conv(2 * n_filter, 4 * n_filter)
        self.encode6 = self.conv(4 * n_filter, 4 * n_filter)
        self.maxpool3 = nn.MaxPool2d(kernel_size=2, stride=2)
        self.encode7 = self.conv(4 * n_filter, 8 * n_filter)
        self.encode8 = self.conv(8 * n_filter, 8 * n_filter)
        self.maxpool4 = nn.MaxPool2d(kernel_size=2, stride=2)

        # Middle
        self.middle_conv1 = self.conv(8 * n_filter, 16 * n_filter)
        self.middle_conv2 = self.conv(16 * n_filter, 16 * n_filter)

        # Decoder
        self.up1 = nn.ConvTranspose2d(16 * n_filter, 8 * n_filter, kernel_size=2, stride=2)
        self.decode1 = self.conv(16 * n_filter, 8 * n_filter)
        self.decode2 = self.conv(8 * n_filter, 8 * n_filter)
        self.up2 = nn.ConvTranspose2d(8 * n_filter, 4 * n_filter, kernel_size=2, stride=2)
        self.decode3 = self.conv(8 * n_filter, 4 * n_filter)
        self.decode4 = self.conv(4 * n_filter, 4 * n_filter)
        self.up3 = nn.ConvTranspose2d(4 * n_filter, 2 * n_filter, kernel_size=2, stride=2)
        self.decode5 = self.conv(4 * n_filter, 2 * n_filter)
        self.decode6 = self.conv(2 * n_filter, 2 * n_filter)
        self.up4 = nn.ConvTranspose2d(2 * n_filter, n_filter, kernel_size=2, stride=2)
        self.decode7 = self.conv(2 * n_filter, n_filter)
        self.decode8 = self.conv(n_filter, n_filter)

        # Output heads
        self.output_layers = nn.ModuleDict()
        for name, config in self.output_heads.items():
            self.output_layers[name] = nn.Conv2d(n_filter, config['channels'], kernel_size=1, padding=0)

    def conv(self, in_channels, out_channels, kernel_size=3, dropout=0., dilation=1):
        layers = [
            nn.Conv2d(in_channels, out_channels, kernel_size, padding=dilation, dilation=dilation),
            nn.BatchNorm2d(out_channels),
            nn.LeakyReLU(negative_slope=0.1, inplace=True),
            nn.Dropout2d(dropout)]
        return nn.Sequential(*layers)

    def concat(self, x1, x2):
        if x1.shape == x2.shape:
            return torch.cat((x1, x2), 1)
        else:
            raise ValueError(f'Concatenation failed: wrong dimensions {x1.shape}, {x2.shape}')

    def apply_activation(self, x, activation):
        if activation == 'sigmoid':
            return torch.sigmoid(x)
        elif activation == 'tanh':
            return torch.tanh(x)
        elif activation == 'relu':
            return torch.relu(x)
        return x

    def forward(self, x):
        # Encoder
        e1 = self.encode1(x)
        e2 = self.encode2(e1)
        m1 = self.maxpool1(e2)
        e3 = self.encode3(m1)
        e4 = self.encode4(e3)
        m2 = self.maxpool2(e4)
        e5 = self.encode5(m2)
        e6 = self.encode6(e5)
        m3 = self.maxpool3(e6)
        e7 = self.encode7(m3)
        e8 = self.encode8(e7)
        m4 = self.maxpool4(e8)

        # Middle
        mid1 = self.middle_conv1(m4)
        mid2 = self.middle_conv2(mid1)

        # Decoder
        u1 = self.up1(mid2)
        c1 = self.concat(u1, e8)
        d1 = self.decode1(c1)
        d2 = self.decode2(d1)
        u2 = self.up2(d2)
        c2 = self.concat(u2, e6)
        d3 = self.decode3(c2)
        d4 = self.decode4(d3)
        u3 = self.up3(d4)
        c3 = self.concat(u3, e4)
        d5 = self.decode5(c3)
        d6 = self.decode6(d5)
        u4 = self.up4(d6)
        c4 = self.concat(u4, e2)
        d7 = self.decode7(c4)
        d8 = self.decode8(d7)

        # Generate outputs for each head
        outputs = {}
        for name, config in self.output_heads.items():
            logits = self.output_layers[name](d8)
            outputs[name] = self.apply_activation(logits, config.get('activation'))

        return outputs



import torch
from torch import nn as nn


class BCEDiceLoss(nn.Module):
    """
    Combines Binary Cross-Entropy Loss and Dice Loss.

    Parameters
    ----------
    alpha : float
        Weight for the Binary Cross-Entropy Loss component.
    beta : float
        Weight for the Dice Loss component.
    """
    def __init__(self, alpha, beta):
        super(BCEDiceLoss, self).__init__()
        self.bce = BCELoss2d()
        self.dice = SoftDiceLoss()
        self.alpha = alpha
        self.beta = beta

    def forward(self, logits, targets):
        """
        Forward pass.

        Parameters
        ----------
        logits : torch.Tensor
            The predicted outputs.
        targets : torch.Tensor
            The ground truth labels.

        Returns
        -------
        torch.Tensor
            The combined loss value.
        """
        return self.alpha * self.bce(logits, targets) + self.beta * self.dice(logits, targets)


class logcoshDiceLoss(nn.Module):
    """
    Log-Cosh Dice Loss.

    Combines the Log-Cosh function with Dice Loss for improved stability.
    """
    def __init__(self):
        super(logcoshDiceLoss, self).__init__()
        self.dice = SoftDiceLoss()

    def forward(self, logits, targets):
        """
        Forward pass.

        Parameters
        ----------
        logits : torch.Tensor
            The predicted outputs.
        targets : torch.Tensor
            The ground truth labels.

        Returns
        -------
        torch.Tensor
            The Log-Cosh Dice loss value.
        """
        x = self.dice(logits, targets)
        return torch.log((torch.exp(x) + torch.exp(-x)) / 2)


class BCELoss2d(nn.Module):
    """
    Binary Cross-Entropy Loss for 2D data.

    Parameters
    ----------
    weight : torch.Tensor, optional
        A manual rescaling weight given to the loss of each batch element.
    reduction : str, optional
        Specifies the reduction to apply to the output: 'none' | 'mean' | 'sum'.
    """
    def __init__(self, weight=None, reduction='mean', **kwargs):
        super(BCELoss2d, self).__init__()
        self.bce_loss = nn.BCELoss(weight, reduction=reduction)

    def forward(self, logits, targets):
        """
        Forward pass.

        Parameters
        ----------
        logits : torch.Tensor
            The predicted outputs.
        targets : torch.Tensor
            The ground truth labels.

        Returns
        -------
        torch.Tensor
            The binary cross-entropy loss value.
        """
        probs = torch.sigmoid(logits)
        probs_flat = probs.view(-1)
        targets_flat = targets.view(-1)
        return self.bce_loss(probs_flat, targets_flat)


class weightedBCELoss(nn.Module):
    """
    Weighted Binary Cross-Entropy Loss.

    Parameters
    ----------
    alpha : float, optional
        Weight for positive examples (default is 1).
    beta : float, optional
        Weight for negative examples (default is 0.1).
    """
    def __init__(self, alpha=1, beta=0.1):
        super(weightedBCELoss, self).__init__()
        self.bce = nn.BCELoss(reduce=False)
        self.alpha, self.beta = alpha, beta

    def forward(self, logits, targets):
        """
        Forward pass.

        Parameters
        ----------
        logits : torch.Tensor
            The predicted outputs.
        targets : torch.Tensor
            The ground truth labels.

        Returns
        -------
        torch.Tensor
            The weighted binary cross-entropy loss value.
        """
        probs = torch.sigmoid(logits)
        # compute weights
        weights = torch.clone(targets)
        weights[targets >= 0.5] = self.alpha
        weights[targets < 0.5] = self.beta
        # compute loss
        loss = torch.mean(self.bce(probs, targets) * weights)
        return loss


class SoftDiceLoss(nn.Module):
    """
    Soft Dice Loss.

    Parameters
    ----------
    smooth : float, optional
        Smoothing factor to avoid division by zero (default is 1.0).
    """
    def __init__(self, smooth=1.0):
        super(SoftDiceLoss, self).__init__()
        self.smooth = smooth

    def forward(self, logits, targets):
        """
        Forward pass.

        Parameters
        ----------
        logits : torch.Tensor
            The predicted outputs.
        targets : torch.Tensor
            The ground truth labels.

        Returns
        -------
        torch.Tensor
            The soft dice loss value.
        """
        num = targets.size(0)
        probs = torch.sigmoid(logits)
        m1 = probs.view(num, -1)
        m2 = targets.view(num, -1)
        intersection = (m1 * m2)

        score = 2. * (intersection.sum(1) + self.smooth) / (m1.sum(1) + m2.sum(1) + self.smooth)
        score = 1 - score.sum() / num
        return score


class TverskyLoss(nn.Module):
    """
    Tversky Loss.

    Parameters
    ----------
    alpha : float, optional
        Weight of false positives (default is 0.5).
    beta : float, optional
        Weight of false negatives (default is 0.5).
    smooth : float, optional
        Smoothing factor to avoid division by zero (default is 1).
    """
    def __init__(self, alpha=0.5, beta=0.5, smooth=1):
        super(TverskyLoss, self).__init__()
        self.alpha = alpha
        self.beta = beta
        self.smooth = smooth

    def forward(self, inputs, targets):
        """
        Forward pass.

        Parameters
        ----------
        inputs : torch.Tensor
            The predicted outputs.
        targets : torch.Tensor
            The ground truth labels.

        Returns
        -------
        torch.Tensor
            The Tversky loss value.
        """
        inputs = torch.sigmoid(inputs)
        inputs = inputs.view(-1)
        targets = targets.view(-1)

        # True Positives, False Positives & False Negatives
        TP = (inputs * targets).sum()
        FP = ((1 - targets) * inputs).sum()
        FN = (targets * (1 - inputs)).sum()

        Tversky = (TP + self.smooth) / (TP + self.alpha * FP + self.beta * FN + self.smooth)

        return 1 - Tversky


class logcoshTverskyLoss(nn.Module):
    """
    Log-Cosh Tversky Loss.

    Parameters
    ----------
    alpha : float, optional
        Weight of false positives (default is 0.5).
    beta : float, optional
        Weight of false negatives (default is 0.5).
    smooth : float, optional
        Smoothing factor to avoid division by zero (default is 1).
    """
    def __init__(self, alpha=0.5, beta=0.5, smooth=1):
        super(logcoshTverskyLoss, self).__init__()
        self.alpha = alpha
        self.beta = beta
        self.smooth = smooth

    def forward(self, inputs, targets):
        """
        Forward pass.

        Parameters
        ----------
        inputs : torch.Tensor
            The predicted outputs.
        targets : torch.Tensor
            The ground truth labels.

        Returns
        -------
        torch.Tensor
            The log-cosh Tversky loss value.
        """
        inputs = torch.sigmoid(inputs)
        inputs = inputs.view(-1)
        targets = targets.view(-1)

        # True Positives, False Positives & False Negatives
        TP = (inputs * targets).sum()
        FP = ((1 - targets) * inputs).sum()
        FN = (targets * (1 - inputs)).sum()

        Tversky = (TP + self.smooth) / (TP + self.alpha * FP + self.beta * FN + self.smooth)

        return torch.log(torch.cosh(1 - Tversky))


import os
import random
import shutil
from pathlib import Path
from typing import List, Tuple, Optional, Callable

import cv2
import numpy as np
import tifffile
import torch
from albumentations import (
    Blur, GaussNoise, ShotNoise,
    RandomBrightnessContrast, Compose, RandomScale, RandomCrop, PadIfNeeded)
from scipy.ndimage import rotate
from torch.utils.data import Dataset
from tqdm import tqdm


class DataProcess(Dataset):
    """Class for training data generation (processing, splitting, augmentation)"""

    def __init__(self,
                 image_dir: str,
                 target_dirs: List[str],
                 target_types: dict[str, str],
                 data_dir: str = '../data/',
                 dim_out: Tuple[int, int] = (256, 256),
                 in_channels: int = 1,
                 out_channels: int = 1,
                 add_tile: int = 0,
                 nan_to_val: float = 0,
                 val_split: float = 0.2,
                 clip_threshold: Tuple[float, float] = (0., 99.99),
                 aug_factor: float = 2,
                 gauss_noise_lims: Tuple[float, float] = (0.01, 0.1),
                 shot_noise_lims: Tuple[float, float] = (0.001, 0.01),
                 brightness_contrast: Tuple[float, float] = (0.1, 0.1),
                 blur_limit: Tuple[int, int] = (3, 5),
                 random_rotate: bool = True,
                 scale_limit: Tuple[float, float] = (0, 0),
                 create: bool = True,
                 file_filter: Optional[Callable[[str], bool]] = None):
        """
        Create training data object for network training

        1) Create folder structure for training data
        2) Move and preprocess training images
        3) Split input images into tiles
        4) Augment training data
        5) Create object of PyTorch Dataset class for training

        Parameters
        ----------
        image_dir : str
            Directory with training images
        target_dirs : List[str]
            List of directories with targets
        target_types : dict
            Dictionary of target types (e.g. {'feature1': 'mask', 'feature2': 'mask'})
        dim_out : Tuple[int, int]
            Resize dimensions of images for training
        data_dir : str
            Base dir of temporary directories for training data
        in_channels : int
            Number of input channels of training data
        out_channels : int
            Number of output channels of training data
        add_tile : int
            Add additional tile for splitting images into tiles with more overlapping tiles
        nan_to_val: float
            Value to use for missing (NaN) pixels in targets
        val_split : float
            Validation split for training
        clip_threshold : Tuple[float, float]
            Clip thresholds for intensity normalization of images
        aug_factor : float
            Augmentation factor, number of patches per image is determined by
            n_pixels_image/n_pixels_patch * aug_factor.
        gauss_noise_lims : Tuple[float, float]
            Limits of Gaussian noise for image augmentation
        shot_noise_lims : Tuple[float, float]
            Limits of Shot noise for image augmentation
        brightness_contrast : Tuple[float, float]
            Adapt brightness and contrast of images during augmentation
        blur_limit : Tuple[float, float]
            Limits of blur (min, max) for image augmentation
        random_rotate : bool
            Whether to randomly rotate images during augmentation
        scale_limit : Tuple[float, float]
            Limits of scale (min, max) for image augmentation
        create : bool, optional
            If False, existing data set in data_path is used
        file_filter : function, optional
            Function with filename as arg that returns bool values indicating whether to include files in the dataset

        Methods
        -------
        __len__()
            Returns the total number of samples.
        __getitem__(idx)
            Retrieves the sample at the specified index.
        """
        self.image_dir = image_dir
        self.target_dirs = target_dirs
        self.target_keys = [os.path.basename(os.path.normpath(_dir)) for _dir in target_dirs]
        self.target_types = target_types
        self.data_dir = data_dir
        self.data = []  # empty list for data dicts
        self.create = create
        self.in_channels = in_channels
        self.out_channels = out_channels
        self.dim_out = dim_out
        self.nan_to_val = nan_to_val
        self.clip_threshold = clip_threshold
        self.add_tile = add_tile
        self.aug_factor = aug_factor
        self.brightness_contrast = brightness_contrast
        self.gauss_noise_lims = gauss_noise_lims
        self.shot_noise_lims = shot_noise_lims
        self.blur_limit = blur_limit
        self.random_rotate = random_rotate
        self.scale_limit = scale_limit
        self.file_filter = file_filter

        self.val_split = val_split
        self.mode = 'train'

        if create:
            # Delete the existing data directory if it exists
            if os.path.exists(self.data_dir):
                shutil.rmtree(self.data_dir)

            # Create a new data directory
            os.makedirs(self.data_dir, exist_ok=True)

            # Proceed with data processing
            self.__read_and_edit()
            self.__augment()

    def __read_and_edit(self):
        # Find all TIFF files in image_dir
        image_path = Path(self.image_dir)
        files_image = [str(file) for ext in ['*.tif', '*.tiff'] for file in image_path.glob(ext)]

        if self.file_filter:
            files_image = [file for file in files_image if self.file_filter(file)]

        # Read and normalize images
        for file_img_i in files_image:
            img_i = tifffile.imread(file_img_i)
            basename_i = os.path.basename(file_img_i)

            # Create dict for image and targets
            data_i = {}

            # Clip and normalize
            img_i = img_i.astype(np.float32)
            img_i = np.clip(img_i, a_min=np.nanpercentile(img_i, self.clip_threshold[0]),
                            a_max=np.percentile(img_i, self.clip_threshold[1]))
            img_i = (img_i - np.nanmin(img_i)) / (np.nanmax(img_i) - np.nanmin(img_i))

            # Add image to dict
            data_i['image'] = img_i

            # Iterate through target folders and add targets to dict
            all_targets_exist = True
            for target_dir_j in self.target_dirs:
                file_target_j = os.path.join(target_dir_j, basename_i)
                if os.path.exists(file_target_j):
                    target_data = tifffile.imread(file_target_j)
                    target_key = os.path.basename(target_dir_j[:-1])  # Use the folder name as the target key
                    data_i[target_key] = target_data
                else:
                    print(f"Warning: Target file {file_target_j} does not exist.")
                    all_targets_exist = False
                    break

            # Add data to the list only if all targets exist and check shape consistency
            if all_targets_exist:
                shapes = []
                for key in data_i:
                    shapes.append(data_i[key].shape)
                    if len(shapes) > 1 and shapes[0] != shapes[-1]:
                        raise ValueError(f"File: {file_img_i}. Shape mismatch: {key} has shape {shapes[-1]} but expected {shapes[0]}")
                self.data.append(data_i)

    def __augment(self):
        target_types = {key: self.target_types[key] for key in self.target_keys}
        aug_pipeline = Compose(transforms=[
            RandomScale(scale_limit=self.scale_limit, p=0.75,
                        interpolation=cv2.INTER_NEAREST,
                        mask_interpolation=cv2.INTER_NEAREST),
            Blur(blur_limit=self.blur_limit, p=0.25),
            PadIfNeeded(min_height=self.dim_out[0], min_width=self.dim_out[1],
                        border_mode=cv2.BORDER_WRAP, position='bottom_left'),
            RandomCrop(height=self.dim_out[0], width=self.dim_out[1], p=1),
            ShotNoise(scale_range=self.shot_noise_lims, p=0.25),
            GaussNoise(std_range=self.gauss_noise_lims, p=0.25),
            RandomBrightnessContrast(brightness_limit=self.brightness_contrast[0],
                                     contrast_limit=self.brightness_contrast[1], p=0.5),
        ], additional_targets=target_types)

        def chw_to_hwc(x):
            if len(x.shape) == 3:  # Only convert if it has channels
                return np.transpose(x, (1, 2, 0))
            return x

        def hwc_to_chw(x):
            if len(x.shape) == 3:  # Only convert if it has channels
                return np.transpose(x, (2, 0, 1))
            return x

        def rotate_array(x, angle, order=1):
            # Convert bool arrays to float32 first
            if x.dtype == bool or x.dtype == np.bool_:
                x = x.astype(np.float32)
                xmin, xmax = 0, 1
                needs_clip = True
            elif np.nanmin(x) >= 0 and np.nanmax(x) <= 1:  # for probability masks with [0, 1] range
                xmin, xmax = np.nanmin(x), np.nanmax(x)
                needs_clip = True
            else:
                xmin, xmax = None, None
                needs_clip = False

            if np.any(np.isnan(x)):
                nan_mask = np.isnan(x)
                x_filled = np.where(nan_mask, 0, x)
                rotated = rotate(x_filled, angle, reshape=False, mode='grid-wrap', order=order)
                rotated_mask = rotate(nan_mask.astype(np.uint8), angle, reshape=False,
                                      mode='grid-wrap', order=order) > 0.5
                rotated = rotated.astype(np.float32)
                rotated[rotated_mask] = np.nan
            else:
                rotated = rotate(x, angle, reshape=False, mode='grid-wrap', order=order)
                rotated = rotated.astype(np.float32)

            # Clip boolean-originated and [0, 1] range arrays
            if needs_clip:
                rotated = np.clip(rotated, xmin, xmax)

            return rotated

        def rotate_array_90(x, factor):
            if len(x.shape) == 3 and x.shape[0] < x.shape[1]:  # CHW format
                return np.rot90(x, factor, axes=(1, 2))
            else:  # HW or HWC format
                return np.rot90(x, factor)

        running_number = 0

        for patch_data in tqdm(self.data):
            # Convert image and targets, only if they have channels
            image_i = patch_data['image'].astype('float32')
            targets_i = {key: patch_data[key].astype('float32')
                         for key in patch_data if key != 'image'}

            aug_factor = max(int((image_i.shape[0] * image_i.shape[1]) /
                             (self.dim_out[0] * self.dim_out[1]) * self.aug_factor), 2)

            for _ in range(aug_factor):
                aug_image = image_i.copy()
                aug_targets = targets_i.copy()
                if self.random_rotate:
                    if random.random() < 0.5:
                        # Apply random rotation with any angle
                        angle = np.random.uniform(0, 360)
                        aug_image = rotate_array(aug_image, angle, order=0)

                        for key in aug_targets:
                            aug_targets[key] = rotate_array(aug_targets[key], angle, order=3)
                            if 'orientation' in key:
                                aug_targets[key] = (aug_targets[key] - np.radians(angle)) % (2 * np.pi)
                    else:
                        # Apply random rotation by 90 degree
                        factor = np.random.randint(0, 3)
                        aug_image = rotate_array_90(aug_image, factor=factor)
                        for key in aug_targets:
                            if 'orientation' in key:
                                # Adjust orientation values and rotate
                                aug_targets[key] = (aug_targets[key] - (np.pi / 2 * factor)) % (2 * np.pi)
                            aug_targets[key] = rotate_array_90(aug_targets[key], factor=factor)

                aug_data_hwc = {'image': np.clip(aug_image, 0, 1)}
                aug_targets_hwc = {key: chw_to_hwc(aug_targets[key]) for key in aug_targets}
                aug_data_hwc.update(aug_targets_hwc)

                augmented = aug_pipeline(**aug_data_hwc)

                # Convert back to CHW format only if needed
                aug_image = hwc_to_chw(augmented['image'])
                aug_targets = {key: hwc_to_chw(augmented[key])
                               for key in targets_i}

                # Save augmented image
                image_dir = os.path.join(self.data_dir, 'image')
                os.makedirs(image_dir, exist_ok=True)
                tifffile.imwrite(os.path.join(image_dir, f'image_{running_number}.tif'),
                                 aug_image)

                # Save augmented targets
                for target_key, target_value in aug_targets.items():
                    target_dir = os.path.join(self.data_dir, target_key)
                    os.makedirs(target_dir, exist_ok=True)
                    tifffile.imwrite(os.path.join(target_dir,
                                                  f'{target_key}_{running_number}.tif'),
                                     target_value)

                running_number += 1

        print(f'Augmentation completed for {running_number} patches.')

    def __len__(self):
        # Count the number of image files in the 'image' directory within the data path
        image_dir = os.path.join(self.data_dir, 'image')
        return len([name for name in os.listdir(image_dir) if os.path.isfile(os.path.join(image_dir, name))])

    def __getitem__(self, idx):
        # Construct the file name for the image
        img_name = f'image_{idx}.tif'

        # Read the image from the 'image' directory
        image_path = os.path.join(self.data_dir, 'image', img_name)
        image_0 = tifffile.imread(image_path)

        # Convert the image to a PyTorch tensor
        image = torch.from_numpy(image_0)

        # Initialize the sample dictionary with the image
        sample = {'image': image}

        # Iterate over each target directory and read the corresponding target file
        for target_dir in self.target_dirs:
            target_key = os.path.basename(os.path.normpath(target_dir))
            target_file = f'{target_key}_{idx}.tif'
            target_path = os.path.join(self.data_dir, target_key, target_file)
            if os.path.exists(target_path):
                target_data = tifffile.imread(target_path).astype('float32')
                # convert to vector field if 'orientation'
                if target_key == 'orientation':
                    target_data = np.stack([np.cos(target_data), np.sin(target_data)])
                # set NaNs to val
                target_data[np.isnan(target_data)] = self.nan_to_val
                sample[target_key] = torch.from_numpy(target_data)
            else:
                raise FileNotFoundError(f"Target file {target_path} not found.")

        # Return the sample as a dictionary
        return sample



import glob
import os
import shutil

import numpy as np
import tifffile
import torch
from albumentations import (
    ShiftScaleRotate, GaussNoise,
    RandomBrightnessContrast, Compose, RandomRotate90)
from skimage import morphology, transform
from torch.utils.data import Dataset


class DataProcess(Dataset):
    """
    Create training data object for network training

    1) Create folder structure for training data
    2) Move and preprocess training images
    3) Split input images into tiles
    4) Augment training data
    5) Create object of PyTorch Dataset class for training

    Parameters
    ----------
    source_dir : Tuple[str, str]
        Path of training data [images, labels]. Images need to be tif files.
    dim_out : Tuple[int, int]
        Resize dimensions of images for training
    aug_factor : int
        Factor of image augmentation
    data_path : str
        Base path of directories for training data
    file_ext : str
        Extension of image files. Defaults to '.tif'
    threshold_masks : int
        Threshold to binarize masks [0-255]
    dilate_mask
        Radius of binary erosion/dilation of masks [-2, -1, 0, 1, 2]
    dilate_kernel : str
        Dilation kernel ('disk' or 'square')
    val_split : float
        Validation split for training
    invert_masks : bool
        If True, greyscale binary labels is inverted
    skeletonize : bool
        If True, binary labels are skeletonized
    create : bool, optional
        If False, existing data set in data_path is used
    clip_threshold : Tuple[float, float]
        Clip thresholds for intensity normalization of images
    shiftscalerotate : [float, float, float]
        Shift, scale and rotate image during augmentation
    noise_amp : float
        Amplitude of Gaussian noise for image augmentation
    brightness_contrast : Tuple[float, float]
        Adapt brightness and contrast of images during augmentation
    rescale : float, optional
        Rescale all images and labels by factor rescale
    """
    def __init__(self, source_dir, dim_out=(256, 256), aug_factor=10, data_path='../data/', file_ext='.tif',
                 threshold_masks=50, dilate_mask=0, dilate_kernel='disk', val_split=0.2, invert_masks=False,
                 skeletonize=False, create=True, clip_threshold=(0.2, 99.8), shiftscalerotate=(0, 0, 0), noise_amp=10,
                 brightness_contrast=(0.25, 0.25), rescale=None):

        self.source_dir = source_dir
        self.create = create
        self.data_path = data_path
        self.file_ext = file_ext
        self.dim_out = dim_out
        self.threshold_masks = threshold_masks
        self.skeletonize = skeletonize
        self.invert_masks = invert_masks
        self.clip_threshold = clip_threshold
        self.aug_factor = aug_factor
        self.shiftscalerotate = shiftscalerotate
        self.brightness_contrast = brightness_contrast
        self.noise_amp = noise_amp
        self.rescale = rescale
        self.dilate_mask = dilate_mask
        self.dilate_kernel = dilate_kernel
        self.val_split = val_split
        self.mode = 'train'

        self.__make_dirs()

        if create:
            self.__move_and_edit()
            self.__merge_images()
            self.__split()
            if self.aug_factor is not None:
                self.__augment()

    def __make_dirs(self):
        self.image_path = self.data_path + '/image/'
        self.prev_image_path = self.data_path + '/prev_image/'
        self.mask_path = self.data_path + '/mask/'
        self.merge_path = self.data_path + '/merge/'
        self.split_merge_path = self.data_path + '/split/merge/'
        self.split_image_path = self.data_path + '/split/image/'
        self.split_prev_image_path = self.data_path + '/split/prev_image/'
        self.split_mask_path = self.data_path + '/split/mask/'
        self.aug_image_path = self.data_path + '/augmentation/aug_image/'
        self.aug_mask_path = self.data_path + '/augmentation/aug_mask/'
        self.aug_prev_image_path = self.data_path + '/augmentation/aug_prev_image/'

        # delete old files
        if self.create:
            if os.path.exists(self.data_path):
                shutil.rmtree(self.data_path)

        # make folders
        os.makedirs(self.image_path, exist_ok=True)
        os.makedirs(self.mask_path, exist_ok=True)
        os.makedirs(self.merge_path, exist_ok=True)
        os.makedirs(self.split_merge_path, exist_ok=True)
        os.makedirs(self.split_image_path, exist_ok=True)
        os.makedirs(self.split_mask_path, exist_ok=True)
        os.makedirs(self.aug_image_path, exist_ok=True)
        os.makedirs(self.aug_mask_path, exist_ok=True)
        os.makedirs(self.aug_prev_image_path, exist_ok=True)
        os.makedirs(self.prev_image_path, exist_ok=True)
        os.makedirs(self.split_prev_image_path, exist_ok=True)

    def __move_and_edit(self):
        # create image data
        files_image = glob.glob(self.source_dir[0] + '*' + self.file_ext)
        for file_i in files_image:
            img_i = tifffile.imread(file_i).astype('float32')
            # clip and normalize (0,255)
            img_i = np.clip(img_i, a_min=np.nanpercentile(img_i, self.clip_threshold[0]),
                            a_max=np.percentile(img_i, self.clip_threshold[1]))
            img_i = (img_i - np.nanmin(img_i)) / (np.nanmax(img_i) - np.nanmin(img_i)) * 255
            if self.rescale is not None:
                img_i = transform.rescale(img_i, self.rescale)
            img_i = img_i.astype('uint8')
            save_i = os.path.splitext(os.path.basename(file_i))[0]
            save_i = save_i.replace(' ', '_')
            # split the input image into two
            # if the previous image and currently image are concatenated horizontally
            if len(img_i.shape) == 2:
                img_width = int(img_i.shape[1] / 2)
                prev_img = img_i[:, :img_width]
                infer_img = img_i[:, img_width:]
            # if the previous image image is stacked as another layer
            elif len(img_i.shape) == 3:
                prev_img = img_i[0]
                infer_img = img_i[1]
            else:
                raise ValueError('Unknown data structure of input images.')
            tifffile.imwrite(self.prev_image_path + save_i + '.tif', prev_img)
            tifffile.imwrite(self.image_path + save_i + '.tif', infer_img)

        # create masks
        files_mask = glob.glob(self.source_dir[1] + '*' + self.file_ext)
        print('%s files found' % len(files_mask))
        for file_i in files_mask:
            mask_i = tifffile.imread(file_i)
            if self.rescale is not None:
                mask_i = transform.rescale(mask_i, self.rescale)
            if self.invert_masks:
                mask_i = 255 - mask_i
            if self.threshold_masks is not None:
                mask_i[mask_i >= self.threshold_masks] = 255
                mask_i[mask_i < self.threshold_masks] = 0
            if self.skeletonize:
                mask_i[mask_i > 1] = 1
                mask_i = morphology.skeletonize(mask_i) * 255
            if self.dilate_kernel == 'disk':
                kernel = morphology.disk
            elif self.dilate_kernel == 'square':
                kernel = morphology.square
            else:
                raise ValueError(f'Dilate kernel {self.dilate_kernel} unknown!')
            if self.dilate_mask > 0:
                mask_i = morphology.dilation(mask_i, kernel(self.dilate_mask))
            elif self.dilate_mask < 0:
                mask_i = morphology.erosion(mask_i, kernel(-self.dilate_mask))
            mask_i = mask_i.astype('uint8')
            save_i = os.path.splitext(os.path.basename(file_i))[0]
            save_i = save_i.replace(' ', '_')
            tifffile.imwrite(self.mask_path + save_i + '.tif', mask_i)

    def __merge_images(self):
        self.mask_files = glob.glob(self.data_path + '/mask/*.tif')
        self.image_files = glob.glob(self.data_path + '/image/*.tif')
        self.prev_image_files = glob.glob(self.data_path + '/prev_image/*.tif')

        if len(self.mask_files) != len(self.image_files):
            raise ValueError('Number of ground truth does not match number of image stacks')

        for i, file_i in enumerate(self.mask_files):
            basename_i = os.path.basename(file_i)
            mask_i = tifffile.imread(self.data_path + '/mask/' + basename_i)
            image_i = tifffile.imread(self.data_path + '/image/' + basename_i)
            prev_image_i = tifffile.imread(self.data_path + '/prev_image/' + basename_i)
            merge = np.zeros((mask_i.shape[0], mask_i.shape[1], 3))
            merge[:, :, 0] = mask_i
            merge[:, :, 1] = image_i
            merge[:, :, 2] = prev_image_i
            merge = merge.astype('uint8')
            tifffile.imwrite(self.merge_path + str(i) + '.tif', merge)

    def __split(self):
        self.merges = glob.glob(self.merge_path + '*.tif')
        n = 0
        for i in range(len(self.merges)):
            merge = tifffile.imread(self.merge_path + str(i) + '.tif')
            dim_in = merge.shape
            # padding if dim_in < dim_out
            x_gap = max(0, self.dim_out[0] - dim_in[0])
            y_gap = max(0, self.dim_out[1] - dim_in[1])
            merge = np.pad(merge, ((0, x_gap), (0, y_gap), (0, 0)), 'reflect')
            # number of patches in x and y
            dim_in = merge.shape
            N_x = int(np.ceil(dim_in[0] / self.dim_out[0]))
            N_y = int(np.ceil(dim_in[1] / self.dim_out[1]))
            # starting indices of patches
            X_start = np.linspace(0, dim_in[0] - self.dim_out[0], N_x).astype('int16')
            Y_start = np.linspace(0, dim_in[1] - self.dim_out[1], N_y).astype('int16')
            for j in range(N_x):
                for k in range(N_y):
                    patch_ij = merge[X_start[j]:X_start[j] + self.dim_out[0], Y_start[k]:Y_start[k] + self.dim_out[1],
                               :]
                    mask_ij = patch_ij[:, :, 0]
                    image_ij = patch_ij[:, :, 1]
                    prev_image_ij = patch_ij[:, :, 2]
                    tifffile.imwrite(self.split_merge_path + f'{n}.tif', patch_ij)
                    tifffile.imwrite(self.split_mask_path + f'{n}.tif', mask_ij)
                    tifffile.imwrite(self.split_image_path + f'{n}.tif', image_ij)
                    tifffile.imwrite(self.split_prev_image_path + f'{n}.tif', prev_image_ij)
                    n += 1

    def __augment(self, p=0.8):
        aug_pipeline = Compose([
            RandomRotate90(p=1.0),
            ShiftScaleRotate(self.shiftscalerotate[0], self.shiftscalerotate[1], self.shiftscalerotate[2]),
            GaussNoise(var_limit=(self.noise_amp, self.noise_amp), p=0.3),
            RandomBrightnessContrast(brightness_limit=self.brightness_contrast[0],
                                     contrast_limit=self.brightness_contrast[1], p=0.5),
        ],
            p=p)

        patches_image = glob.glob(self.split_image_path + '*.tif')
        patches_mask = glob.glob(self.split_mask_path + '*.tif')
        patches_prev_image = glob.glob(self.split_prev_image_path + '*.tif')
        n_patches = len(patches_image)
        k = 0
        for i in range(n_patches):
            image_i = tifffile.imread(patches_image[i])
            mask_i = tifffile.imread(patches_mask[i])
            prev_image_i = tifffile.imread(patches_prev_image[i])
            data_i = {'image': np.dstack([image_i, prev_image_i]), 'mask': mask_i}
            data_aug_i = np.asarray([aug_pipeline(**data_i) for _ in range(self.aug_factor)])
            imgs_aug_i = np.asarray([data_aug_i[j]['image'][:, :, 0] for j in range(self.aug_factor)])
            masks_aug_i = np.asarray([data_aug_i[j]['mask'] for j in range(self.aug_factor)])
            prev_image_aug_i = np.asarray([data_aug_i[j]['image'][:, :, 1] for j in range(self.aug_factor)])

            for j in range(self.aug_factor):
                tifffile.imwrite(self.aug_image_path + f'{k}.tif', imgs_aug_i[j])
                tifffile.imwrite(self.aug_mask_path + f'{k}.tif', masks_aug_i[j])
                tifffile.imwrite(self.aug_prev_image_path + f'{k}.tif', prev_image_aug_i[j])
                k += 1
        print(f'Number of training images: {k}', )

    def __len__(self):
        if self.aug_factor is not None:
            return len(os.listdir(self.aug_image_path))
        else:
            return len(os.listdir(self.split_image_path))

    def __getitem__(self, idx):
        imgname = str(idx) + '.tif'
        midname = os.path.basename(imgname)
        if self.aug_factor is not None:
            image_0 = tifffile.imread(self.aug_image_path + midname).astype('float32') / 255
            mask_0 = tifffile.imread(self.aug_mask_path + midname).astype('float32') / 255
            prev_image_0 = tifffile.imread(self.aug_prev_image_path + midname).astype('float32') / 255
        else:
            image_0 = tifffile.imread(self.split_image_path + midname).astype('float32') / 255
            mask_0 = tifffile.imread(self.split_image_path + midname).astype('float32') / 255
            prev_image_0 = tifffile.imread(self.split_prev_image_path + midname).astype('float32') / 255
        image = torch.from_numpy(image_0)
        mask = torch.from_numpy(mask_0)
        prev_image = torch.from_numpy(prev_image_0)
        del image_0, mask_0, prev_image_0
        sample = {'image': image, 'mask': mask, 'prev_image': prev_image}
        return sample
